@@ -73,6 +73,11 @@ static void redraw_custom_statusline(win_T *wp);
 static int  did_update_one_window;
 #endif
 
+#ifdef FEAT_EVAL
+static void redraw_listener_cleanup(void);
+static void invoke_redraw_listener_start_or_end(bool start);
+#endif
+
 /*
  * Based on the current value of curwin->w_topline, transfer a screenfull
  * of stuff from Filemem to ScreenLines[], and update curwin->w_botline.
@@ -209,6 +214,9 @@ update_screen(int type_arg)
 		    redraw_cmdline = TRUE;
 		redraw_tabline = TRUE;
 	    }
+#if defined(FEAT_TABPANEL)
+	    redraw_tabpanel = TRUE;
+#endif
 	}
 	msg_scrolled = 0;
 	need_wait_return = FALSE;
@@ -241,6 +249,10 @@ update_screen(int type_arg)
 	curwin->w_redr_type = UPD_NOT_VALID;
 #endif
 
+#ifdef FEAT_EVAL
+    invoke_redraw_listener_start_or_end(true);
+#endif
+
     // Only start redrawing if there is really something to do.
     if (type == UPD_INVERTED)
 	update_curswant();
@@ -265,6 +277,11 @@ update_screen(int type_arg)
     if (redraw_tabline || type >= UPD_NOT_VALID)
 	draw_tabline();
 
+#if defined(FEAT_TABPANEL)
+    if (redraw_tabpanel || type >= UPD_NOT_VALID)
+	draw_tabpanel();
+#endif
+
 #ifdef FEAT_SYN_HL
     // Correct stored syntax highlighting info for changes in each displayed
     // buffer.  Each buffer must only be done once.
@@ -397,6 +414,12 @@ update_screen(int type_arg)
 	gui_update_scrollbars(FALSE);
     }
 #endif
+
+#ifdef FEAT_EVAL
+    invoke_redraw_listener_start_or_end(false);
+    redraw_listener_cleanup();
+#endif
+
     return OK;
 }
 
@@ -425,11 +448,8 @@ statusline_row(win_T *wp)
 win_redr_status(win_T *wp, int ignore_pum UNUSED)
 {
     int		row;
-    char_u	*p;
-    int		len;
     int		fillchar;
     int		attr;
-    int		this_ru_col;
     static int  busy = FALSE;
 
     // It's possible to get here recursively when 'statusline' (indirectly)
@@ -463,11 +483,17 @@ win_redr_status(win_T *wp, int ignore_pum UNUSED)
 #endif
     else
     {
+	char_u	*p;
+	int	plen;
+	int	NameBufflen;
+	int	this_ru_col;
+	int	n;			// scratch value
+
 	fillchar = fillchar_status(&attr, wp);
 
 	get_trans_bufname(wp->w_buffer);
 	p = NameBuff;
-	len = (int)STRLEN(p);
+	plen = (int)STRLEN(p);
 
 	if ((bt_help(wp->w_buffer)
 #ifdef FEAT_QUICKFIX
@@ -475,74 +501,63 @@ win_redr_status(win_T *wp, int ignore_pum UNUSED)
 #endif
 		    || bufIsChanged(wp->w_buffer)
 		    || wp->w_buffer->b_p_ro)
-		&& len < MAXPATHL - 1)
-	    *(p + len++) = ' ';
-	if (bt_help(wp->w_buffer))
+		&& plen < MAXPATHL - 1)
 	{
-	    vim_snprintf((char *)p + len, MAXPATHL - len, "%s", _("[Help]"));
-	    len += (int)STRLEN(p + len);
+	    *(p + plen++) = ' ';	// replace NUL with space
+	    *(p + plen) = NUL;		// NUL terminate the string
 	}
+	if (bt_help(wp->w_buffer))
+	    plen += vim_snprintf((char *)p + plen, MAXPATHL - plen, "%s", _("[Help]"));
 #ifdef FEAT_QUICKFIX
 	if (wp->w_p_pvw)
-	{
-	    vim_snprintf((char *)p + len, MAXPATHL - len, "%s", _("[Preview]"));
-	    len += (int)STRLEN(p + len);
-	}
+	    plen += vim_snprintf((char *)p + plen, MAXPATHL - plen, "%s", _("[Preview]"));
 #endif
 	if (bufIsChanged(wp->w_buffer) && !bt_terminal(wp->w_buffer))
-	{
-	    vim_snprintf((char *)p + len, MAXPATHL - len, "%s", "[+]");
-	    len += (int)STRLEN(p + len);
-	}
+	    plen += vim_snprintf((char *)p + plen, MAXPATHL - plen, "%s", "[+]");
 	if (wp->w_buffer->b_p_ro)
-	{
-	    vim_snprintf((char *)p + len, MAXPATHL - len, "%s", _("[RO]"));
-	    len += (int)STRLEN(p + len);
-	}
+	    plen += vim_snprintf((char *)p + plen, MAXPATHL - plen, "%s", _("[RO]"));
 
-	this_ru_col = ru_col - (Columns - wp->w_width);
-	if (this_ru_col < (wp->w_width + 1) / 2)
-	    this_ru_col = (wp->w_width + 1) / 2;
+	this_ru_col = ru_col - (cmdline_width - wp->w_width);
+	n = (wp->w_width + 1) / 2;
+	if (this_ru_col < n)
+	    this_ru_col = n;
 	if (this_ru_col <= 1)
 	{
 	    p = (char_u *)"<";		// No room for file name!
-	    len = 1;
+	    plen = 1;
 	}
 	else if (has_mbyte)
 	{
-	    int	clen = 0, i;
+	    int	i;
 
 	    // Count total number of display cells.
-	    clen = mb_string2cells(p, -1);
+	    plen = mb_string2cells(p, -1);
 
 	    // Find first character that will fit.
 	    // Going from start to end is much faster for DBCS.
-	    for (i = 0; p[i] != NUL && clen >= this_ru_col - 1;
+	    for (i = 0; p[i] != NUL && plen >= this_ru_col - 1;
 		    i += (*mb_ptr2len)(p + i))
-		clen -= (*mb_ptr2cells)(p + i);
-	    len = clen;
+		plen -= (*mb_ptr2cells)(p + i);
 	    if (i > 0)
 	    {
 		p = p + i - 1;
 		*p = '<';
-		++len;
+		++plen;
 	    }
-
 	}
-	else if (len > this_ru_col - 1)
+	else if (plen > this_ru_col - 1)
 	{
-	    p += len - (this_ru_col - 1);
+	    p += plen - (this_ru_col - 1);
 	    *p = '<';
-	    len = this_ru_col - 1;
+	    plen = this_ru_col - 1;
 	}
 
 	screen_puts(p, row, wp->w_wincol, attr);
-	screen_fill(row, row + 1, len + wp->w_wincol,
+	screen_fill(row, row + 1, plen + wp->w_wincol,
 			this_ru_col + wp->w_wincol, fillchar, fillchar, attr);
-
-	if (get_keymap_str(wp, (char_u *)"<%s>", NameBuff, MAXPATHL)
-		&& (this_ru_col - len) > (int)(STRLEN(NameBuff) + 1))
-	    screen_puts(NameBuff, row, (int)(this_ru_col - STRLEN(NameBuff)
+	if ((NameBufflen = get_keymap_str(wp, (char_u *)"<%s>", NameBuff, MAXPATHL)) > 0
+		&& (this_ru_col - plen) > (NameBufflen + 1))
+	    screen_puts(NameBuff, row, (int)(this_ru_col - NameBufflen
 						   - 1 + wp->w_wincol), attr);
 
 	win_redr_ruler(wp, TRUE, ignore_pum);
@@ -550,7 +565,8 @@ win_redr_status(win_T *wp, int ignore_pum UNUSED)
 	// Draw the 'showcmd' information if 'showcmdloc' == "statusline".
 	if (p_sc && *p_sloc == 's')
 	{
-	    int	width = MIN(10, this_ru_col - len - 2);
+	    n = this_ru_col - plen - 2;		    // perform the calculation here so we only do it once
+	    int	width = MIN(10, n);
 
 	    if (width > 0)
 		screen_puts_len(showcmd_buf, width, row,
@@ -626,24 +642,17 @@ showruler(int always)
     // Redraw the tab pages line if needed.
     if (redraw_tabline)
 	draw_tabline();
+
+#if defined(FEAT_TABPANEL)
+    if (redraw_tabpanel)
+	draw_tabpanel();
+#endif
 }
 
     void
 win_redr_ruler(win_T *wp, int always, int ignore_pum)
 {
-#define RULER_BUF_LEN 70
-    char_u	buffer[RULER_BUF_LEN];
-    int		row;
-    int		fillchar;
-    int		attr;
-    int		empty_line = FALSE;
-    colnr_T	virtcol;
-    int		i;
-    size_t	len;
-    int		o;
-    int		this_ru_col;
-    int		off = 0;
-    int		width;
+    int	empty_line = FALSE;
 
     // If 'ruler' off don't do anything
     if (!p_ru)
@@ -661,6 +670,7 @@ win_redr_ruler(win_T *wp, int always, int ignore_pum)
     if (wp == lastwin && lastwin->w_status_height == 0)
 	if (edit_submode != NULL)
 	    return;
+
     // Don't draw the ruler when the popup menu is visible, it may overlap.
     // Except when the popup menu will be redrawn anyway.
     if (!ignore_pum && pum_visible())
@@ -698,6 +708,21 @@ win_redr_ruler(win_T *wp, int always, int ignore_pum)
 #endif
 	    || empty_line != wp->w_ru_empty)
     {
+	int	row;
+	int	fillchar;
+	int	attr;
+	int	off;
+	int	width;
+	colnr_T	virtcol;
+#define RULER_BUF_LEN 70
+	char_u	buffer[RULER_BUF_LEN];
+	int	bufferlen;
+	char_u	rel_pos[RULER_BUF_LEN];
+	int	rel_poslen;
+	int	this_ru_col;
+	int	n1;			    // scratch value
+	int	n2;			    // scratch value
+
 	cursor_off();
 	if (wp->w_status_height)
 	{
@@ -711,8 +736,8 @@ win_redr_ruler(win_T *wp, int always, int ignore_pum)
 	    row = Rows - 1;
 	    fillchar = ' ';
 	    attr = 0;
-	    width = Columns;
-	    off = 0;
+	    off = cmdline_col_off;
+	    width = cmdline_width;
 	}
 
 	// In list mode virtcol needs to be recomputed
@@ -724,16 +749,14 @@ win_redr_ruler(win_T *wp, int always, int ignore_pum)
 	    wp->w_p_list = TRUE;
 	}
 
-	/*
-	 * Some sprintfs return the length, some return a pointer.
-	 * To avoid portability problems we use strlen() here.
-	 */
-	vim_snprintf((char *)buffer, RULER_BUF_LEN, "%ld,",
+	// row number, column number is appended
+	// l10n: leave as-is unless a space after the comma is preferred
+	// l10n: do not add any row/column label, due to the limited space
+	bufferlen = vim_snprintf((char *)buffer, RULER_BUF_LEN, _("%ld,"),
 		(wp->w_buffer->b_ml.ml_flags & ML_EMPTY)
 		    ? 0L
 		    : (long)(wp->w_cursor.lnum));
-	len = STRLEN(buffer);
-	col_print(buffer + len, RULER_BUF_LEN - len,
+	bufferlen += col_print(buffer + bufferlen, RULER_BUF_LEN - bufferlen,
 			empty_line ? 0 : (int)wp->w_cursor.col + 1,
 			(int)virtcol + 1);
 
@@ -742,56 +765,60 @@ win_redr_ruler(win_T *wp, int always, int ignore_pum)
 	 * On the last line, don't print in the last column (scrolls the
 	 * screen up on some terminals).
 	 */
-	i = (int)STRLEN(buffer);
-	get_rel_pos(wp, buffer + i + 1, RULER_BUF_LEN - i - 1);
-	o = i + vim_strsize(buffer + i + 1);
+	rel_poslen = get_rel_pos(wp, rel_pos, RULER_BUF_LEN);
+	n1 = bufferlen + vim_strsize(rel_pos);
 	if (wp->w_status_height == 0)	// can't use last char of screen
-	    ++o;
-	this_ru_col = ru_col - (Columns - width);
-	if (this_ru_col < 0)
-	    this_ru_col = 0;
+	    ++n1;
+
+	this_ru_col = ru_col - (cmdline_width - width);
 	// Never use more than half the window/screen width, leave the other
 	// half for the filename.
-	if (this_ru_col < (width + 1) / 2)
-	    this_ru_col = (width + 1) / 2;
-	if (this_ru_col + o < width)
+	n2 = (width + 1) / 2;
+	if (this_ru_col < n2)
+	    this_ru_col = n2;
+	if (this_ru_col + n1 < width)
 	{
-	    // need at least 3 chars left for get_rel_pos() + NUL
-	    while (this_ru_col + o < width && RULER_BUF_LEN > i + 4)
+	    // need at least space for rel_pos + NUL
+	    while (this_ru_col + n1 < width
+		    && RULER_BUF_LEN > bufferlen + rel_poslen + 1)	// +1 for NUL
 	    {
 		if (has_mbyte)
-		    i += (*mb_char2bytes)(fillchar, buffer + i);
+		    bufferlen += (*mb_char2bytes)(fillchar, buffer + bufferlen);
 		else
-		    buffer[i++] = fillchar;
-		++o;
+		    buffer[bufferlen++] = fillchar;
+		++n1;
 	    }
-	    get_rel_pos(wp, buffer + i, RULER_BUF_LEN - i);
+	    bufferlen += vim_snprintf((char *)buffer + bufferlen, RULER_BUF_LEN - bufferlen,
+			    "%s", rel_pos);
 	}
 	// Truncate at window boundary.
 	if (has_mbyte)
 	{
-	    o = 0;
-	    for (i = 0; buffer[i] != NUL; i += (*mb_ptr2len)(buffer + i))
+	    for (n1 = 0, n2 = 0; buffer[n1] != NUL; n1 += (*mb_ptr2len)(buffer + n1))
 	    {
-		o += (*mb_ptr2cells)(buffer + i);
-		if (this_ru_col + o > width)
+		n2 += (*mb_ptr2cells)(buffer + n1);
+		if (this_ru_col + n2 > width)
 		{
-		    buffer[i] = NUL;
+		    bufferlen = n1;
+		    buffer[bufferlen] = NUL;
 		    break;
 		}
 	    }
 	}
-	else if (this_ru_col + (int)STRLEN(buffer) > width)
-	    buffer[width - this_ru_col] = NUL;
+	else if (this_ru_col + bufferlen > width)
+	{
+	    bufferlen = width - this_ru_col;
+	    buffer[bufferlen] = NUL;
+	}
 
 	screen_puts(buffer, row, this_ru_col + off, attr);
-	i = redraw_cmdline;
+	n1 = redraw_cmdline;
 	screen_fill(row, row + 1,
-		this_ru_col + off + (int)STRLEN(buffer),
+		this_ru_col + off + bufferlen,
 		(off + width),
 		fillchar, fillchar, attr);
 	// don't redraw the cmdline because of showing the ruler
-	redraw_cmdline = i;
+	redraw_cmdline = n1;
 	wp->w_ru_cursor = wp->w_cursor;
 	wp->w_ru_virtcol = wp->w_virtcol;
 	wp->w_ru_empty = empty_line;
@@ -948,9 +975,10 @@ text_to_screenline(win_T *wp, char_u *text, int col)
     else
     {
 	int len = (int)STRLEN(text);
+	int n = wp->w_width - col;
 
-	if (len > wp->w_width - col)
-	    len = wp->w_width - col;
+	if (len > n)
+	    len = n;
 	if (len > 0)
 	{
 #ifdef FEAT_RIGHTLEFT
@@ -1028,11 +1056,12 @@ redraw_win_toolbar(win_T *wp)
     }
     wp->w_winbar_items[item_idx].wb_menu = NULL; // end marker
 
-    screen_line(wp, wp->w_winrow, wp->w_wincol, wp->w_width, wp->w_width, 0);
+    screen_line(wp, wp->w_winrow, wp->w_wincol, wp->w_width,
+							  wp->w_width, -1, 0);
 }
 #endif
 
-#if defined(FEAT_FOLDING) || defined(PROTO)
+#if defined(FEAT_FOLDING)
 /*
  * Copy "buf[len]" to ScreenLines["off"] and set attributes to "attr".
  */
@@ -1087,7 +1116,7 @@ fold_line(
 
     // 1. Add the cmdwin_type for the command-line window
     // Ignores 'rightleft', this window is never right-left.
-    if (cmdwin_type != 0 && wp == curwin)
+    if (wp == cmdwin_win)
     {
 	ScreenLines[off] = cmdwin_type;
 	ScreenAttrs[off] = HL_ATTR(HLF_AT);
@@ -1216,7 +1245,7 @@ fold_line(
 		}
 	    }
 
-	    sprintf((char *)buf, fmt, w, num);
+	    vim_snprintf((char *)buf, sizeof(buf), fmt, w, num);
 #ifdef FEAT_RIGHTLEFT
 	    if (wp->w_p_rl)
 		// the line number isn't reversed
@@ -1300,7 +1329,7 @@ fold_line(
 			&& (lnume < bot->lnum
 			    || (lnume == bot->lnum
 				&& (bot->col - (*p_sel == 'e'))
-		>= (colnr_T)STRLEN(ml_get_buf(wp->w_buffer, lnume, FALSE)))))))
+				    >= ml_get_buf_len(wp->w_buffer, lnume))))))
 	{
 	    if (VIsual_mode == Ctrl_V)
 	    {
@@ -1362,8 +1391,8 @@ fold_line(
     }
 #endif
 
-    screen_line(wp, row + W_WINROW(wp), wp->w_wincol,
-						  wp->w_width, wp->w_width, 0);
+    screen_line(wp, row + W_WINROW(wp), wp->w_wincol, wp->w_width, wp->w_width,
+	    -1, 0);
 
     // Update w_cline_height and w_cline_folded if the cursor line was
     // updated (saves a call to plines() later).
@@ -1408,11 +1437,11 @@ fold_line(
  *		   - if wp->w_buffer->b_mod_set set, update lines between
  *		     b_mod_top and b_mod_bot.
  *		   - if wp->w_redraw_top non-zero, redraw lines between
- *		     wp->w_redraw_top and wp->w_redr_bot.
+ *		     wp->w_redraw_top and wp->w_redraw_bot.
  *		   - continue redrawing when syntax status is invalid.
  *		4. if scrolled up, update lines at the bottom.
  * This results in three areas that may need updating:
- * top:	from first row to top_end (when scrolled down)
+ * top: from first row to top_end (when scrolled down)
  * mid: from mid_start to mid_end (update inversion or changed text)
  * bot: from bot_start to last row (when scrolled up)
  */
@@ -1431,6 +1460,8 @@ win_update(win_T *wp)
 				// updating.  999 when no bot area updating
     int		scrolled_down = FALSE;	// TRUE when scrolled down when
 					// w_topline got smaller a bit
+    int		scrolled_for_mod = FALSE;   // TRUE after scrolling for changed
+					    // lines
 #ifdef FEAT_SEARCH_EXTRA
     int		top_to_mod = FALSE;    // redraw above mod_top
 #endif
@@ -1537,7 +1568,7 @@ win_update(win_T *wp)
 
     // Make sure skipcol is valid, it depends on various options and the window
     // width.
-    if (wp->w_skipcol > 0)
+    if (wp->w_skipcol > 0 && wp->w_width > win_col_off(wp))
     {
 	int w = 0;
 	int width1 = wp->w_width - win_col_off(wp);
@@ -1567,14 +1598,6 @@ win_update(win_T *wp)
     else
 #endif
 
-    if (buf->b_mod_set && buf->b_mod_xlines != 0 && wp->w_redraw_top != 0)
-    {
-	// When there are both inserted/deleted lines and specific lines to be
-	// redrawn, w_redraw_top and w_redraw_bot may be invalid, just redraw
-	// everything (only happens when redrawing is off for while).
-	type = UPD_NOT_VALID;
-    }
-    else
     {
 	// Set mod_top to the first line that needs displaying because of
 	// changes.  Set mod_bot to the first line after the changes.
@@ -1699,11 +1722,6 @@ win_update(win_T *wp)
 		top_end = 1;
 #endif
 	}
-
-	// When line numbers are displayed need to redraw all lines below
-	// inserted/deleted lines.
-	if (mod_top != 0 && buf->b_mod_xlines != 0 && wp->w_p_nu)
-	    mod_bot = MAXLNUM;
     }
     wp->w_redraw_top = 0;	// reset for next time
     wp->w_redraw_bot = 0;
@@ -1793,7 +1811,7 @@ win_update(win_T *wp)
 	    if (j < wp->w_height - 2)		// not too far off
 	    {
 		i = plines_m_win(wp, wp->w_topline, wp->w_lines[0].wl_lnum - 1,
-									 TRUE);
+								wp->w_height);
 #ifdef FEAT_DIFF
 		// insert extra lines for previously invisible filler lines
 		if (wp->w_lines[0].wl_lnum != wp->w_topline)
@@ -1918,9 +1936,8 @@ win_update(win_T *wp)
 		    // Correct the first entry for filler lines at the top
 		    // when it won't get updated below.
 		    if (wp->w_p_diff && bot_start > 0)
-			wp->w_lines[0].wl_size =
-			    plines_win_nofill(wp, wp->w_topline, TRUE)
-							      + wp->w_topfill;
+			wp->w_lines[0].wl_size = plines_correct_topline(wp,
+							  wp->w_topline, TRUE);
 #endif
 		}
 	    }
@@ -2056,8 +2073,7 @@ win_update(win_T *wp)
 			{
 			    colnr_T t;
 
-			    pos.col = (int)STRLEN(ml_get_buf(wp->w_buffer,
-							     pos.lnum, FALSE));
+			    pos.col = (int)ml_get_buf_len(wp->w_buffer, pos.lnum);
 			    getvvcol(wp, &pos, NULL, NULL, &t);
 			    if (toc < t)
 				toc = t;
@@ -2263,6 +2279,7 @@ win_update(win_T *wp)
 				// match in fixed position might need redraw
 				// if lines were inserted or deleted
 				|| (wp->w_match_head != NULL
+						    && buf->b_mod_set
 						    && buf->b_mod_xlines != 0)
 #endif
 				))))
@@ -2280,13 +2297,15 @@ win_update(win_T *wp)
 	    // When at start of changed lines: May scroll following lines
 	    // up or down to minimize redrawing.
 	    // Don't do this when the change continues until the end.
-	    // Don't scroll when dollar_vcol >= 0, keep the "$".
-	    // Don't scroll when redrawing the top, scrolled already above.
-	    if (lnum == mod_top
-		    && mod_bot != MAXLNUM
-		    && !(dollar_vcol >= 0 && mod_bot == mod_top + 1)
-		    && row >= top_end)
+	    // Don't scroll for changed lines in the top area if that's already
+	    // done above, but do scroll for changed lines below the top area.
+	    if (!scrolled_for_mod && mod_bot != MAXLNUM
+		    && lnum >= mod_top && lnum < MAX(mod_bot, mod_top + 1)
+		    && (!scrolled_down || row >= top_end))
 	    {
+		scrolled_for_mod = TRUE;
+
+		int		old_cline_height = 0;
 		int		old_rows = 0;
 		int		new_rows = 0;
 		int		xtra_rows;
@@ -2302,6 +2321,8 @@ win_update(win_T *wp)
 		    if (wp->w_lines[i].wl_valid
 			    && wp->w_lines[i].wl_lnum == mod_bot)
 			break;
+		    if (wp->w_lines[i].wl_lnum == wp->w_cursor.lnum)
+			old_cline_height = wp->w_lines[i].wl_size;
 		    old_rows += wp->w_lines[i].wl_size;
 #ifdef FEAT_FOLDING
 		    if (wp->w_lines[i].wl_valid
@@ -2332,26 +2353,17 @@ win_update(win_T *wp)
 		    j = idx;
 		    for (l = lnum; l < mod_bot; ++l)
 		    {
+			if (dollar_vcol >= 0 && wp == curwin &&
+				old_cline_height > 0 && l == wp->w_cursor.lnum)
+			    // When dollar_vcol >= 0, cursor line isn't fully
+			    // redrawn, and its height remains unchanged.
+			    new_rows += old_cline_height;
 #ifdef FEAT_FOLDING
-			if (hasFoldingWin(wp, l, NULL, &l, TRUE, NULL))
+			else if (hasFoldingWin(wp, l, NULL, &l, TRUE, NULL))
 			    ++new_rows;
+#endif
 			else
-#endif
-			{
-#ifdef FEAT_DIFF
-			    if (l == wp->w_topline)
-			    {
-				int n = plines_win_nofill(wp, l, FALSE)
-								+ wp->w_topfill;
-				n -= adjust_plines_for_skipcol(wp);
-				if (n > wp->w_height)
-				    n = wp->w_height;
-				new_rows += n;
-			    }
-			    else
-#endif
-				new_rows += plines_win(wp, l, TRUE);
-			}
+			    new_rows += plines_correct_topline(wp, l, TRUE);
 			++j;
 			if (new_rows > wp->w_height - row - 2)
 			{
@@ -2503,7 +2515,7 @@ win_update(win_T *wp)
 #endif
 
 		// Display one line.
-		row = win_line(wp, lnum, srow, wp->w_height, FALSE, &spv);
+		row = win_line(wp, lnum, srow, wp->w_height, 0, &spv);
 
 #ifdef FEAT_FOLDING
 		wp->w_lines[idx].wl_folded = FALSE;
@@ -2518,18 +2530,20 @@ win_update(win_T *wp)
 	    wp->w_lines[idx].wl_lnum = lnum;
 	    wp->w_lines[idx].wl_valid = TRUE;
 
+	    int is_curline = wp == curwin && lnum == wp->w_cursor.lnum;
+
 	    // Past end of the window or end of the screen. Note that after
 	    // resizing wp->w_height may be end up too big. That's a problem
 	    // elsewhere, but prevent a crash here.
 	    if (row > wp->w_height || row + wp->w_winrow >= Rows)
 	    {
 		// we may need the size of that too long line later on
-		if (dollar_vcol == -1)
+		if (dollar_vcol == -1 || !is_curline)
 		    wp->w_lines[idx].wl_size = plines_win(wp, lnum, TRUE);
 		++idx;
 		break;
 	    }
-	    if (dollar_vcol == -1)
+	    if (dollar_vcol == -1 || !is_curline)
 		wp->w_lines[idx].wl_size = row - srow;
 	    ++idx;
 #ifdef FEAT_FOLDING
@@ -2540,17 +2554,23 @@ win_update(win_T *wp)
 	}
 	else
 	{
-	    if (wp->w_p_rnu && wp->w_last_cursor_lnum_rnu != wp->w_cursor.lnum)
+	    // If:
+	    // - 'number' is set and below inserted/deleted lines, or
+	    // - 'relativenumber' is set and cursor moved vertically,
+	    // the text doesn't need to be redrawn, but the number column does.
+	    if ((wp->w_p_nu && mod_top != 0 && lnum >= mod_bot
+			&& buf->b_mod_set && buf->b_mod_xlines != 0)
+		    || (wp->w_p_rnu
+			&& wp->w_last_cursor_lnum_rnu != wp->w_cursor.lnum))
 	    {
 #ifdef FEAT_FOLDING
-		// 'relativenumber' set and the cursor moved vertically: The
-		// text doesn't need to be drawn, but the number column does.
 		fold_count = foldedCount(wp, lnum, &win_foldinfo);
 		if (fold_count != 0)
 		    fold_line(wp, fold_count, &win_foldinfo, lnum, row);
 		else
 #endif
-		    (void)win_line(wp, lnum, srow, wp->w_height, TRUE, &spv);
+		    (void)win_line(wp, lnum, srow, wp->w_height,
+					       wp->w_lines[idx].wl_size, &spv);
 	    }
 
 	    // This line does not need to be drawn, advance to the next one.
@@ -2600,13 +2620,16 @@ win_update(win_T *wp)
 
 	for (k = 0; k < Rows; ++k)
 	    if (enc_utf8)
-		if ((*mb_off2cells)(LineOffset[k] + Columns - 2,
+		if ((*mb_off2cells)(LineOffset[k] + topframe->fr_width - 2,
 					   LineOffset[k] + screen_Columns) > 1)
-		    screen_draw_rectangle(k, Columns - 2, 1, 2, FALSE);
+		    screen_draw_rectangle(k, topframe->fr_width - 2, 1, 2,
+			    FALSE);
 		else
-		    screen_draw_rectangle(k, Columns - 1, 1, 1, FALSE);
+		    screen_draw_rectangle(k, topframe->fr_width - 1, 1, 1,
+			    FALSE);
 	    else
-		screen_char(LineOffset[k] + Columns - 1, k, Columns - 1);
+		screen_char(LineOffset[k] + topframe->fr_width - 1, k,
+			cmdline_width - 1);
     }
 #endif
 
@@ -2676,7 +2699,7 @@ win_update(win_T *wp)
 	    // Last line isn't finished: Display "@@@" at the end.
 	    screen_fill(W_WINROW(wp) + wp->w_height - 1,
 		    W_WINROW(wp) + wp->w_height,
-		    start_col < wp->w_wincol ? wp->w_wincol : start_col,
+		    (start_col < wp->w_wincol ? wp->w_wincol : start_col),
 		    (int)W_ENDCOL(wp),
 		    symbol, symbol, HL_ATTR(HLF_AT));
 	    set_empty_rows(wp, srow);
@@ -2711,7 +2734,7 @@ win_update(win_T *wp)
 	    }
 #endif
 	}
-	else if (dollar_vcol == -1)
+	else if (dollar_vcol == -1 || wp != curwin)
 	    wp->w_botline = lnum;
 
 	// Make sure the rest of the screen is blank.
@@ -2736,7 +2759,7 @@ win_update(win_T *wp)
     wp->w_old_botfill = wp->w_botfill;
 #endif
 
-    if (dollar_vcol == -1)
+    if (dollar_vcol == -1 || wp != curwin)
     {
 	// There is a trick with w_botline.  If we invalidate it on each
 	// change that might modify it, this will cause a lot of expensive
@@ -2780,7 +2803,7 @@ win_update(win_T *wp)
 		    redrawWinline(wp, wp->w_cursor.lnum);
 	    }
 #endif
-	    // New redraw either due to updated topline or due to wcol fix.
+	    // New redraw either due to updated topline, wcol fix or reset skipcol.
 	    if (wp->w_redr_type != 0)
 	    {
 		// Don't update for changes in buffer again.
@@ -2788,6 +2811,7 @@ win_update(win_T *wp)
 		curbuf->b_mod_set = FALSE;
 		j = curbuf->b_mod_xlines;
 		curbuf->b_mod_xlines = 0;
+		curs_columns(TRUE);
 		win_update(curwin);
 		curbuf->b_mod_set = i;
 		curbuf->b_mod_xlines = j;
@@ -2860,7 +2884,7 @@ update_finish(void)
 }
 #endif
 
-#if defined(FEAT_NETBEANS_INTG) || defined(PROTO)
+#if defined(FEAT_NETBEANS_INTG)
     void
 update_debug_sign(buf_T *buf, linenr_T lnum)
 {
@@ -2901,11 +2925,16 @@ update_debug_sign(buf_T *buf, linenr_T lnum)
 	    win_redr_status(wp, FALSE);
     }
 
+#if defined(FEAT_TABPANEL)
+    if (redraw_tabpanel)
+	draw_tabpanel();
+#endif
+
     update_finish();
 }
 #endif
 
-#if defined(FEAT_GUI) || defined(PROTO)
+#if defined(FEAT_GUI)
 /*
  * Update a single window, its status line and maybe the command line msg.
  * Used for the GUI scrollbar.
@@ -2933,6 +2962,11 @@ updateWindow(win_T *wp)
     if (redraw_tabline)
 	draw_tabline();
 
+#if defined(FEAT_TABPANEL)
+    if (redraw_tabpanel)
+	draw_tabpanel();
+#endif
+
     if (wp->w_redr_status || p_ru
 # ifdef FEAT_STL_OPT
 	    || *p_stl != NUL || *wp->w_p_stl != NUL
@@ -3056,7 +3090,7 @@ redraw_asap(int type)
 		    mch_memmove(ScreenLines2 + off,
 				screenline2 + r * cols,
 				(size_t)cols * sizeof(schar_T));
-		screen_line(curwin, cmdline_row + r, 0, cols, cols, 0);
+		screen_line(curwin, cmdline_row + r, 0, cols, cols, -1, 0);
 	    }
 	    ret = 4;
 	}
@@ -3166,6 +3200,13 @@ redraw_win_later(
     win_T	*wp,
     int		type)
 {
+#ifdef FEAT_EVAL
+    // If inside a redraw_listener_add() callback, then only set the redraw type
+    // for the window and not request another one right after.
+    if (inside_redraw_on_start_cb && wp->w_redr_type < type)
+	wp->w_redr_type = type;
+    else
+#endif
     if (!exiting && !redraw_not_allowed && wp->w_redr_type < type)
     {
 	wp->w_redr_type = type;
@@ -3222,7 +3263,11 @@ redraw_all_windows_later(int type)
     void
 set_must_redraw(int type)
 {
-    if (!redraw_not_allowed && must_redraw < type)
+    if (!redraw_not_allowed &&
+#ifdef FEAT_EVAL
+	    !inside_redraw_on_start_cb &&
+#endif
+	    must_redraw < type)
 	must_redraw = type;
 }
 
@@ -3252,7 +3297,7 @@ redraw_buf_later(buf_T *buf, int type)
 #endif
 }
 
-#if defined(FEAT_SIGNS) || defined(PROTO)
+#if defined(FEAT_SIGNS)
     void
 redraw_buf_line_later(buf_T *buf, linenr_T lnum)
 {
@@ -3265,7 +3310,7 @@ redraw_buf_line_later(buf_T *buf, linenr_T lnum)
 }
 #endif
 
-#if defined(FEAT_JOB_CHANNEL) || defined(PROTO)
+#if defined(FEAT_JOB_CHANNEL)
     void
 redraw_buf_and_status_later(buf_T *buf, int type)
 {
@@ -3331,6 +3376,11 @@ redraw_statuslines(void)
 	    win_redr_status(wp, FALSE);
     if (redraw_tabline)
 	draw_tabline();
+
+#if defined(FEAT_TABPANEL)
+    if (redraw_tabpanel)
+	draw_tabpanel();
+#endif
 }
 
 /*
@@ -3368,9 +3418,205 @@ redrawWinline(
     win_T	*wp,
     linenr_T	lnum)
 {
-    if (wp->w_redraw_top == 0 || wp->w_redraw_top > lnum)
-	wp->w_redraw_top = lnum;
-    if (wp->w_redraw_bot == 0 || wp->w_redraw_bot < lnum)
-	wp->w_redraw_bot = lnum;
-    redraw_win_later(wp, UPD_VALID);
+    redraw_win_range_later(wp, lnum, lnum);
 }
+
+    void
+redraw_win_range_later(
+    win_T	*wp,
+    linenr_T	first,
+    linenr_T	last)
+{
+    if (last >= wp->w_topline && first < wp->w_botline)
+    {
+	if (wp->w_redraw_top == 0 || wp->w_redraw_top > first)
+	    wp->w_redraw_top = first;
+	if (wp->w_redraw_bot == 0 || wp->w_redraw_bot < last)
+	    wp->w_redraw_bot = last;
+	redraw_win_later(wp, UPD_VALID);
+    }
+}
+
+#ifdef FEAT_EVAL
+static bool redraw_cb_in_progress = false;
+
+    void
+f_redraw_listener_add(typval_T *argvars, typval_T *rettv)
+{
+    redraw_listener_T	*rln;
+    dict_T		*dict;
+    typval_T		tv;
+    bool		got_one = false;
+    static int		id;
+
+    if (redraw_cb_in_progress)
+    {
+	emsg(_(e_cannot_add_redraw_listener_in_listener_callback));
+	return;
+    }
+
+    if (check_for_dict_arg(argvars, 0) == FAIL)
+	return;
+
+    rln = ALLOC_CLEAR_ONE(redraw_listener_T);
+
+    if (rln == NULL)
+	return;
+    dict = argvars[0].vval.v_dict;
+
+    /*
+     * on_start: called on each screen redraw
+     *
+     * on_end: called at the end of a redraw cycle
+     */
+    if (dict_get_tv(dict, "on_start", &tv) == OK)
+    {
+	callback_T cb = get_callback(&tv);
+
+	if (cb.cb_name == NULL)
+	{
+	    clear_tv(&tv);
+	    vim_free(rln);
+	    return;
+	}
+	copy_callback(&rln->rl_callbacks.on_start, &cb);
+	free_callback(&cb);
+	clear_tv(&tv);
+	got_one = true;
+    }
+
+    if (dict_get_tv(dict, "on_end", &tv) == OK)
+    {
+	callback_T cb = get_callback(&tv);
+
+	if (cb.cb_name == NULL)
+	{
+	    clear_tv(&tv);
+	    free_callback(&rln->rl_callbacks.on_start);
+	    vim_free(rln);
+	    return;
+	}
+	copy_callback(&rln->rl_callbacks.on_end, &cb);
+	free_callback(&cb);
+	clear_tv(&tv);
+	got_one = true;
+    }
+
+    if (!got_one)
+    {
+	emsg(_(e_no_redraw_listener_callbacks_defined));
+	vim_free(rln);
+	return;
+    }
+
+    rln->rl_next = redraw_listeners;
+    redraw_listeners = rln;
+    rln->rl_id = ++id; // Never zero
+
+    rettv->v_type = VAR_NUMBER;
+    rettv->vval.v_number = id;
+
+    return;
+}
+
+    static void
+redraw_listener_free(redraw_listener_T *rln)
+{
+    free_callback(&rln->rl_callbacks.on_start);
+    free_callback(&rln->rl_callbacks.on_end);
+
+    vim_free(rln);
+}
+
+
+    static void
+redraw_listener_cleanup(void)
+{
+    for (redraw_listener_T *rln = redraw_listeners; rln != NULL;)
+    {
+	redraw_listener_T *next = rln->rl_next;
+	if (rln->rl_id == 0)
+	{
+	    if (redraw_listeners == rln)
+		redraw_listeners = rln->rl_next;
+	    redraw_listener_free(rln);
+	}
+	rln = next;
+    }
+}
+
+/*
+ * Return the redraw listener struct with the specified id. Returns NULL if not
+ * found.
+ */
+    static redraw_listener_T *
+get_redraw_listener(int id)
+{
+    for (redraw_listener_T *rln = redraw_listeners; rln != NULL; rln = rln->rl_next)
+	if (rln->rl_id == id)
+	    return rln;
+    return NULL;
+}
+
+    void
+f_redraw_listener_remove(typval_T *argvars, typval_T *rettv UNUSED)
+{
+    int			id;
+    redraw_listener_T	*rln;
+
+    if (check_for_number_arg(argvars, 0) == FAIL)
+	return;
+
+    id = argvars[0].vval.v_number;
+    rln = get_redraw_listener(id);
+
+    rettv->v_type = VAR_NUMBER;
+    if (rln == NULL)
+    {
+	rettv->vval.v_number = 0;
+	return;
+    }
+
+    // We set the id to zero instead of freeing it here, since we still need
+    // rl_next from it.
+    rln->rl_id = 0;
+    rettv->vval.v_number = 1;
+}
+
+/*
+ * Invoke the on_start callbacks.
+ */
+    static void
+invoke_redraw_listener_start_or_end(bool start)
+{
+    typval_T argv[1];
+    typval_T rettv;
+
+    argv[0].v_type = VAR_UNKNOWN;
+
+    if (start)
+	inside_redraw_on_start_cb = true;
+
+    redraw_cb_in_progress = true;
+    for (redraw_listener_T *rln = redraw_listeners; rln != NULL; rln = rln->rl_next)
+    {
+	if (rln->rl_id == 0)
+	    // Listener has been removed, skip
+	    continue;
+	if (start && rln->rl_callbacks.on_start.cb_name != NULL)
+	{
+	    call_callback(&rln->rl_callbacks.on_start, -1, &rettv, 0, argv);
+	    clear_tv(&rettv);
+	}
+	else if (rln->rl_callbacks.on_end.cb_name != NULL)
+	{
+	    call_callback(&rln->rl_callbacks.on_end, -1, &rettv, 0, argv);
+	    clear_tv(&rettv);
+	}
+    }
+    redraw_cb_in_progress = false;
+
+    if (start)
+	inside_redraw_on_start_cb = false;
+}
+#endif // FEAT_EVAL
