@@ -2025,7 +2025,7 @@ EXTERN listitem_T range_list_item;
 EXTERN evalarg_T EVALARG_EVALUATE
 # ifdef DO_INIT
 	= {EVAL_EVALUATE, 0, NULL, NULL, NULL, NULL, GA_EMPTY, GA_EMPTY, NULL,
-			 {0, 0, (int)sizeof(char_u *), 20, NULL}, 0, NULL}
+			 {0, 0, (int)sizeof(char_u *), 20, NULL}, 0, NULL, NULL}
 # endif
 	;
 #endif
@@ -2134,7 +2134,7 @@ INIT(= CLIENTSERVER_METHOD_NONE);
 EXTERN char_u *client_socket INIT(= NULL);
 #endif
 
-// If the <xOSC> key should be propogated from vgetc()
+// If the <xOSC> key should be propagated from vgetc()
 EXTERN int allow_osc_key INIT(= 0);
 
 #ifdef FEAT_EVAL
