@@ -232,9 +232,7 @@ do_mouse(
     int		moved;		// Has cursor moved?
     int		in_status_line;	// mouse in status line
     static int	in_tab_line = FALSE; // mouse clicked in tab line
-#if defined(FEAT_TABPANEL)
     static int	in_tabpanel = FALSE; // mouse clicked in tabpanel
-#endif
     int		in_sep_line;	// mouse in vertical separator line
     int		c1, c2;
 #if defined(FEAT_FOLDING)
@@ -337,7 +335,11 @@ do_mouse(
 
     // Ignore drag and release events if we didn't get a click.
     if (is_click)
+    {
 	got_click = TRUE;
+	in_tab_line = FALSE;
+	in_tabpanel = FALSE;
+    }
     else
     {
 	if (!got_click)			// didn't get click, ignore
@@ -345,16 +347,10 @@ do_mouse(
 	if (!is_drag)			// release, reset got_click
 	{
 	    got_click = FALSE;
-	    if (in_tab_line
-#if defined(FEAT_TABPANEL)
-		|| in_tabpanel
-#endif
-		    )
+	    if (in_tab_line || in_tabpanel)
 	    {
 		in_tab_line = FALSE;
-#if defined(FEAT_TABPANEL)
 		in_tabpanel = FALSE;
-#endif
 		return FALSE;
 	    }
 	}
@@ -479,36 +475,66 @@ do_mouse(
 
     start_visual.lnum = 0;
 
-    // Check for clicking in the tab page line.
+    struct tabpage_label_info {
+	bool is_panel;	    // label type. true: tabpanel, false: tab line
+	bool just_in;	    // just in tabpage label area
+	bool just_click;    // just click tabpage label area
+	int nr;		    // tabpage number
+    } tp_label = { false, false, false, 0 };
+
+    // Check for clicking in the tab page panel.
 #if defined(FEAT_TABPANEL)
-    if (mouse_col < TPL_LCOL(NULL))
+    if (mouse_col < firstwin->w_wincol
+		|| mouse_col >= firstwin->w_wincol + topframe->fr_width)
+    {
+	tp_label.is_panel = true;
+	tp_label.just_in = true;
+	tp_label.nr = get_tabpagenr_on_tabpanel();
+
+	// click in a tab selects that tab page
+	if (is_click && cmdwin_type == 0)
+	    tp_label.just_click = true;
+    }
+    else
+#endif
+    // Check for clicking in the tab page line.
+    if (TabPageIdxs != NULL && mouse_row == 0 && firstwin->w_winrow > 0)
+    {
+	tp_label.just_in = true;
+	tp_label.nr = TabPageIdxs[mouse_col];
+
+	// click in a tab selects that tab page
+	if (is_click && cmdwin_type == 0
+		&& mouse_col < firstwin->w_wincol + topframe->fr_width)
+	    tp_label.just_click = true;
+    }
+
+    if (tp_label.just_in)
     {
 	if (is_drag)
 	{
-	    if (in_tabpanel)
+	    if (in_tabpanel || in_tab_line)
 	    {
-		c1 = get_tabpagenr_on_tabpanel();
+		c1 = tp_label.nr;
 		tabpage_move(c1 <= 0 ? 9999 : c1 < tabpage_index(curtab)
-								? c1 - 1 : c1);
+							    ? c1 - 1 : c1);
 	    }
 	    return FALSE;
 	}
 
-	// click in a tab selects that tab page
-	if (is_click
-# ifdef FEAT_CMDWIN
-		&& cmdwin_type == 0
-# endif
-		&& mouse_col < Columns)
+	if (tp_label.just_click)
 	{
-	    in_tabpanel = TRUE;
-	    c1 = get_tabpagenr_on_tabpanel();
+	    if (tp_label.is_panel)
+		in_tabpanel = TRUE;
+	    else
+		in_tab_line = TRUE;
+	    c1 = tp_label.nr;
 	    if (c1 >= 0)
 	    {
 		if ((mod_mask & MOD_MASK_MULTI_CLICK) == MOD_MASK_2CLICK)
 		{
 		    // double click opens new page
-		    end_visual_mode();
+		    end_visual_mode_keep_button();
 		    tabpage_new();
 		    tabpage_move(c1 == 0 ? 9999 : c1 - 1);
 		}
@@ -520,7 +546,7 @@ do_mouse(
 
 		    // It's like clicking on the status line of a window.
 		    if (curwin != old_curwin)
-			end_visual_mode();
+			end_visual_mode_keep_button();
 		}
 	    }
 	    else
@@ -543,82 +569,17 @@ do_mouse(
 	}
 	return TRUE;
     }
-    else if (is_drag && in_tabpanel)
+    else if (is_drag && (in_tabpanel || (in_tab_line && TabPageIdxs != NULL)))
     {
-	c1 = get_tabpagenr_on_tabpanel();
+#if defined(FEAT_TABPANEL)
+	if (in_tabpanel)
+	    c1 = get_tabpagenr_on_tabpanel();
+	else
+#endif
+	    c1 = TabPageIdxs[mouse_col];
 	tabpage_move(c1 <= 0 ? 9999 : c1 - 1);
 	return FALSE;
     }
-#endif
-
-    if (TabPageIdxs != NULL)  // only when initialized
-    {
-	// Check for clicking in the tab page line.
-	if (mouse_row == 0 && firstwin->w_winrow > 0)
-	{
-	    if (is_drag)
-	    {
-		if (in_tab_line)
-		{
-		    c1 = TabPageIdxs[mouse_col];
-		    tabpage_move(c1 <= 0 ? 9999 : c1 < tabpage_index(curtab)
-								? c1 - 1 : c1);
-		}
-		return FALSE;
-	    }
-
-	    // click in a tab selects that tab page
-	    if (is_click && cmdwin_type == 0 && mouse_col < Columns)
-	    {
-		in_tab_line = TRUE;
-		c1 = TabPageIdxs[mouse_col];
-		if (c1 >= 0)
-		{
-		    if ((mod_mask & MOD_MASK_MULTI_CLICK) == MOD_MASK_2CLICK)
-		    {
-			// double click opens new page
-			end_visual_mode_keep_button();
-			tabpage_new();
-			tabpage_move(c1 == 0 ? 9999 : c1 - 1);
-		    }
-		    else
-		    {
-			// Go to specified tab page, or next one if not clicking
-			// on a label.
-			goto_tabpage(c1);
-
-			// It's like clicking on the status line of a window.
-			if (curwin != old_curwin)
-			    end_visual_mode_keep_button();
-		    }
-		}
-		else
-		{
-		    tabpage_T	*tp;
-
-		    // Close the current or specified tab page.
-		    if (c1 == -999)
-			tp = curtab;
-		    else
-			tp = find_tabpage(-c1);
-		    if (tp == curtab)
-		    {
-			if (first_tabpage->tp_next != NULL)
-			    tabpage_close(FALSE);
-		    }
-		    else if (tp != NULL)
-			tabpage_close_other(tp, FALSE);
-		}
-	    }
-	    return TRUE;
-	}
-	else if (is_drag && in_tab_line)
-	{
-	    c1 = TabPageIdxs[mouse_col];
-	    tabpage_move(c1 <= 0 ? 9999 : c1 - 1);
-	    return FALSE;
-	}
-    }
 
     // When 'mousemodel' is "popup" or "popup_setpos", translate mouse events:
     // right button up   -> pop-up menu
@@ -1525,7 +1486,7 @@ set_mouse_termcode(
 	has_mouse_termcode |= HMT_NORMAL;
 }
 
-#if defined(UNIX) || defined(VMS) || defined(PROTO)
+#if defined(UNIX) || defined(VMS)
     void
 del_mouse_termcode(
     int		n)	// KS_MOUSE, KS_NETTERM_MOUSE or KS_DEC_MOUSE
@@ -1725,10 +1686,6 @@ jump_to_mouse(
     int		mouse_char = ' ';
 #endif
 
-    col -= TPL_LCOL(NULL);
-    if (col < 0)
-	return IN_TABPANEL;
-
     mouse_past_bottom = FALSE;
     mouse_past_eol = FALSE;
 
@@ -1813,7 +1770,7 @@ retnomove:
 
     if (!(flags & MOUSE_FOCUS))
     {
-	if (row < 0 || col + TPL_LCOL(NULL) < 0) // check if it makes sense
+	if (row < 0 || col < 0) // check if it makes sense
 	    return IN_UNKNOWN;
 
 	// find the window where the row is in and adjust "row" and "col" to be
@@ -2052,7 +2009,7 @@ retnomove:
 	    for (first = TRUE; curwin->w_topline > 1; )
 	    {
 #ifdef FEAT_DIFF
-		if (curwin->w_topfill < diff_check(curwin, curwin->w_topline))
+		if (curwin->w_topfill < diff_check_fill(curwin, curwin->w_topline))
 		    ++count;
 		else
 #endif
@@ -2064,7 +2021,7 @@ retnomove:
 		(void)hasFolding(curwin->w_topline, &curwin->w_topline, NULL);
 #endif
 #ifdef FEAT_DIFF
-		if (curwin->w_topfill < diff_check(curwin, curwin->w_topline))
+		if (curwin->w_topfill < diff_check_fill(curwin, curwin->w_topline))
 		    ++curwin->w_topfill;
 		else
 #endif
@@ -2151,7 +2108,8 @@ retnomove:
 #ifdef FEAT_FOLDING
 	// Remember the character under the mouse, it might be a '-' or '+' in
 	// the fold column.
-	mouse_char = ScreenLines[off];
+	mouse_char = enc_utf8 && ScreenLinesUC[off] != 0
+				       ? ScreenLinesUC[off] : ScreenLines[off];
 #endif
     }
 
@@ -2358,6 +2316,10 @@ check_termcode_mouse(
 	 *	   0x23 = any button release
 	 *	   0x60 = button 4 down (scroll wheel down)
 	 *	   0x61 = button 5 down (scroll wheel up)
+	 *	   0x62 = button 6 down (scroll wheel left)
+	 *	   0x63 = button 7 down (scroll wheel right)
+	 *	   0xa0 = button 8 down (backward button)
+	 *	   0xa1 = button 9 down (forward button)
 	 *	add 0x04 for SHIFT
 	 *	add 0x08 for ALT
 	 *	add 0x10 for CTRL
@@ -2512,7 +2474,7 @@ check_termcode_mouse(
 	 * Linux console with GPM and the MS-DOS or Win32 console
 	 * (multi-clicks use >= 0x60).
 	 */
-	if (mouse_code >= MOUSEWHEEL_LOW
+	if (mouse_code >= MOUSEWHEEL_LOW && mouse_code < MOUSESIDEBUTTONS_LOW
 #  ifdef FEAT_GUI
 		&& !gui.in_use
 #  endif
@@ -3035,7 +2997,13 @@ check_termcode_mouse(
 	held_button = MOUSE_RELEASE;
     }
     else
+    {
+#if defined(UNIX)
+	if (use_xterm_mouse() && orig_mouse_code >= MOUSESIDEBUTTONS_LOW)
+	    current_button = (current_button) ? MOUSE_X2 : MOUSE_X1;
+#endif
 	key_name[1] = get_pseudo_mouse_code(current_button, is_click, is_drag);
+    }
 
 
     // Make sure the mouse position is valid.  Some terminals may return weird
@@ -3220,7 +3188,14 @@ mouse_find_win(int *rowp, int *colp, mouse_find_T popup UNUSED)
 #endif
 
     fp = topframe;
+
+    if (*colp < firstwin->w_wincol
+	    || *colp >= firstwin->w_wincol + fp->fr_width
+	    || *rowp < firstwin->w_winrow)
+	return NULL;
+
     *rowp -= firstwin->w_winrow;
+    *colp -= firstwin->w_wincol;
     for (;;)
     {
 	if (fp->fr_layout == FR_LEAF)
@@ -3258,7 +3233,7 @@ mouse_find_win(int *rowp, int *colp, mouse_find_T popup UNUSED)
 }
 
 #if defined(NEED_VCOL2COL) || defined(FEAT_BEVAL) || defined(FEAT_PROP_POPUP) \
-	|| defined(FEAT_EVAL) || defined(PROTO)
+	|| defined(FEAT_EVAL)
 /*
  * Convert a virtual (screen) column to a character column.
  * The first column is zero.
@@ -3288,7 +3263,7 @@ vcol2col(win_T *wp, linenr_T lnum, int vcol, colnr_T *coladdp)
 }
 #endif
 
-#if defined(FEAT_EVAL) || defined(PROTO)
+#if defined(FEAT_EVAL)
 /*
  * "getmousepos()" function.
  */
@@ -3333,9 +3308,6 @@ f_getmousepos(typval_T *argvars UNUSED, typval_T *rettv)
 	    winid = wp->w_id;
 	    winrow = row + 1;
 	    wincol = col + 1;
-	    wincol -= TPL_LCOL(NULL);
-	    if (wincol < 0)
-		wincol = 0;
 	    row -= top_off;
 	    col -= left_off;
 	    if (row >= 0 && row < wp->w_height && col >= 0 && col < wp->w_width)
