@@ -23,9 +23,6 @@
 #undef TRUE		// will be redefined by exec/types.h
 #undef FALSE
 
-// cproto fails on missing include files, skip them
-#ifndef PROTO
-
 #ifndef LATTICE
 # include <exec/exec.h>
 # include <intuition/intuition.h>
@@ -57,8 +54,6 @@
 # include <libraries/arp_pragmas.h>
 #endif
 
-#endif // PROTO
-
 /*
  * Set stack size to 1 MiB on NG systems. This should be enough even for
  * hungry syntax HL / plugin combinations. Leave the stack alone on OS 3
@@ -302,9 +297,7 @@ mch_init(void)
 #endif
 }
 
-#ifndef PROTO
-# include <workbench/startup.h>
-#endif
+#include <workbench/startup.h>
 
 /*
  * Check_win checks whether we have an interactive window.
@@ -1040,9 +1033,7 @@ mch_settmode(tmode_T tmode)
  * Heavely modified by mool.
  */
 
-#ifndef PROTO
-# include <devices/conunit.h>
-#endif
+#include <devices/conunit.h>
 
 /*
  * Get console size in a system friendly way on AROS and MorphOS.
@@ -1215,11 +1206,9 @@ out_num(long n)
  * say 'oml lib:amiga.lib -r sendpacket.o'
  */
 
-#ifndef PROTO
-// #include <proto/exec.h>
-// #include <proto/dos.h>
-# include <exec/memory.h>
-#endif
+//#include <proto/exec.h>
+//#include <proto/dos.h>
+#include <exec/memory.h>
 
 /*
  * Function - dos_packet written by Phil Lindsay, Carolyn Scheppner, and Andy
@@ -1768,3 +1757,31 @@ mch_setenv(char *var, char *value, int x UNUSED)
 	return 0;   // success
     return -1;	    // failure
 }
+
+/*
+ * Fill the buffer 'buf' with 'len' random bytes.
+ * Returns FAIL if RANDOM: is not available or something went wrong.
+ */
+    int
+mch_get_random(char_u *buf, int len)
+{
+    struct Process *proc = (struct Process *) FindTask(0L);
+    APTR win = proc->pr_WindowPtr;
+
+    // Don't show requester if RANDOM: doesn't exist
+    proc->pr_WindowPtr = (APTR) -1L;
+
+    BPTR fh = Open("RANDOM:", MODE_OLDFILE);
+
+    proc->pr_WindowPtr = win;
+
+    int status;
+
+    if (!fh || Read(fh, buf, len) != len)
+	status = FAIL;
+    else
+	status = OK;
+
+    Close(fh);
+    return status;
+}
