@@ -36,7 +36,8 @@
 
 #include "wayland.h"
 
-vwl_connection_T *wayland_ct;
+vwl_connection_T    *wayland_ct;
+bool		    is_reading = false;
 
 /*
  * Like wl_display_flush but always writes all the data in the buffer to the
@@ -677,6 +678,7 @@ wayland_uninit_connection(void)
 #endif
     vwl_connection_destroy(wayland_ct);
     wayland_ct = NULL;
+    is_reading = false;
     wayland_set_display("");
 }
 
@@ -713,6 +715,12 @@ wayland_prepare_read(void)
     if (wayland_ct == NULL)
 	return -1;
 
+    if (is_reading)
+    {
+	wl_display_cancel_read(wayland_ct->display.proxy);
+	is_reading = false;
+    }
+
     while (wl_display_prepare_read(wayland_ct->display.proxy) == -1)
 	// Event queue not empty, dispatch the events
 	if (wl_display_dispatch_pending(wayland_ct->display.proxy) == -1)
@@ -723,11 +731,14 @@ wayland_prepare_read(void)
 	wl_display_cancel_read(wayland_ct->display.proxy);
 	return -1;
     }
+
+    is_reading = true;
+
     return wayland_ct->display.fd;
 }
 
 /*
- * Catch up on any qeueued events
+ * Catch up on any queued events
  */
     int
 wayland_update(void)
@@ -737,7 +748,7 @@ wayland_update(void)
     return vwl_connection_roundtrip(wayland_ct);
 }
 
-#ifndef HAVE_SELECT
+#if !defined(HAVE_SELECT)
 
     void
 wayland_poll_check(int revents)
@@ -745,12 +756,15 @@ wayland_poll_check(int revents)
     if (wayland_ct == NULL)
 	return;
 
+    is_reading = false;
     if (revents & POLLIN)
+    {
 	if (wl_display_read_events(wayland_ct->display.proxy) != -1)
 	{
 	    wl_display_dispatch_pending(wayland_ct->display.proxy);
 	    return;
 	}
+    }
     else if (revents & (POLLHUP | POLLERR))
 	wl_display_cancel_read(wayland_ct->display.proxy);
     else
@@ -762,7 +776,8 @@ wayland_poll_check(int revents)
     wayland_restore_connection();
 }
 
-#else // ifdef HAVE_SELECT
+#endif
+#if defined(HAVE_SELECT)
 
     void
 wayland_select_check(bool is_set)
@@ -770,18 +785,22 @@ wayland_select_check(bool is_set)
     if (wayland_ct == NULL)
 	return;
 
+    is_reading = false;
     if (is_set)
     {
 	if (wl_display_read_events(wayland_ct->display.proxy) != -1)
 	    wl_display_dispatch_pending(wayland_ct->display.proxy);
 	else
+	{
+	    wl_display_cancel_read(wayland_ct->display.proxy);
 	    wayland_restore_connection();
+	}
     }
     else
 	wl_display_cancel_read(wayland_ct->display.proxy);
 }
 
-#endif // !HAVE_SELECT
+#endif
 
 /*
  * Disconnect then reconnect Wayland connection, and update clipmethod.
@@ -830,7 +849,6 @@ ex_wlrestore(exarg_T *eap)
 	smsg(_("restoring Wayland display %s"), wayland_display_name);
 
 #ifdef FEAT_WAYLAND_CLIPBOARD
-	did_warn_clipboard = FALSE;
 	clip_init_wayland();
 #endif
     }
@@ -842,11 +860,11 @@ ex_wlrestore(exarg_T *eap)
     choose_clipmethod();
 }
 
-#ifdef FEAT_WAYLAND_CLIPBOARD
+#if defined(FEAT_WAYLAND_CLIPBOARD)
 
 /*
  * Get a suitable data device manager from connection. "supported" should be
- * iniitialized to VWL_DATA_PROTOCOL_NONE beforehand. Returns NULL if there are
+ * initialized to VWL_DATA_PROTOCOL_NONE beforehand. Returns NULL if there are
  * no data device manager available with the required selection.
  */
     vwl_data_device_manager_T *
